# Copyright 2023 Agnostiq Inc.

"""Functions to convert lattice -> LatticeSchema"""

from covalent._workflow.lattice import Lattice

from ..shared.schemas.lattice import (
    ASSET_FILENAME_MAP,
    LatticeAssets,
    LatticeMetadata,
    LatticeSchema,
)
from .common import AssetType, load_asset, save_asset
from .transport_graph import deserialize_transport_graph, serialize_transport_graph

ASSET_TYPES = {
    "workflow_function": AssetType.TRANSPORTABLE,
    "workflow_function_string": AssetType.TEXT,
    "doc": AssetType.TEXT,
    "named_args": AssetType.OBJECT,
    "named_kwargs": AssetType.OBJECT,
    "cova_imports": AssetType.OBJECT,
    "lattice_imports": AssetType.OBJECT,
    "deps": AssetType.OBJECT,
    "call_before": AssetType.OBJECT,
    "call_after": AssetType.OBJECT,
}


def _serialize_lattice_metadata(lat) -> LatticeMetadata:
    name = lat.__name__
    executor = lat.metadata["executor"]
    executor_data = lat.metadata["executor_data"]
    workflow_executor = lat.metadata["workflow_executor"]
    workflow_executor_data = lat.metadata["workflow_executor_data"]
    return LatticeMetadata(
        name=name,
        executor=executor,
        executor_data=executor_data,
        workflow_executor=workflow_executor,
        workflow_executor_data=workflow_executor_data,
    )


def _deserialize_lattice_metadata(meta: LatticeMetadata) -> dict:
    return {
        "__name__": meta.name,
        "metadata": {
            "executor": meta.executor,
            "executor_data": meta.executor_data,
            "workflow_executor": meta.workflow_executor,
            "workflow_executor_data": meta.workflow_executor_data,
        },
    }


def _serialize_lattice_assets(lat, storage_path: str) -> LatticeAssets:
    workflow_func_asset = save_asset(
        lat.workflow_function,
        ASSET_TYPES["workflow_function"],
        storage_path,
        ASSET_FILENAME_MAP["workflow_function"],
    )

    try:
        workflow_func_string = lat.workflow_function_string
    except AttributeError:
        workflow_func_string = ""
    workflow_func_str_asset = save_asset(
        workflow_func_string,
        ASSET_TYPES["workflow_function_string"],
        storage_path,
        ASSET_FILENAME_MAP["workflow_function_string"],
    )

    docstring = "" if lat.__doc__ is None else lat.__doc__
    docstring_asset = save_asset(
        docstring,
        ASSET_TYPES["doc"],
        storage_path,
        ASSET_FILENAME_MAP["doc"],
    )

    # Deprecate
    named_args_asset = save_asset(
        lat.named_args,
        ASSET_TYPES["named_args"],
        storage_path,
        ASSET_FILENAME_MAP["named_args"],
    )
    named_kwargs_asset = save_asset(
        lat.named_kwargs,
        ASSET_TYPES["named_kwargs"],
        storage_path,
        ASSET_FILENAME_MAP["named_kwargs"],
    )
    cova_imports_asset = save_asset(
        lat.cova_imports,
        ASSET_TYPES["cova_imports"],
        storage_path,
        ASSET_FILENAME_MAP["cova_imports"],
    )
    lattice_imports_asset = save_asset(
        lat.lattice_imports,
        ASSET_TYPES["lattice_imports"],
        storage_path,
        ASSET_FILENAME_MAP["lattice_imports"],
    )

    # NOTE: these are actually JSONable
    deps_asset = save_asset(
        lat.metadata["deps"],
        ASSET_TYPES["deps"],
        storage_path,
        ASSET_FILENAME_MAP["deps"],
    )
    call_before_asset = save_asset(
        lat.metadata["call_before"],
        ASSET_TYPES["call_before"],
        storage_path,
        ASSET_FILENAME_MAP["call_before"],
    )
    call_after_asset = save_asset(
        lat.metadata["call_after"],
        ASSET_TYPES["call_after"],
        storage_path,
        ASSET_FILENAME_MAP["call_after"],
    )

    return LatticeAssets(
        workflow_function=workflow_func_asset,
        workflow_function_string=workflow_func_str_asset,
        doc=docstring_asset,
        named_args=named_args_asset,
        named_kwargs=named_kwargs_asset,
        cova_imports=cova_imports_asset,
        lattice_imports=lattice_imports_asset,
        deps=deps_asset,
        call_before=call_before_asset,
        call_after=call_after_asset,
    )


def _deserialize_lattice_assets(assets: LatticeAssets) -> dict:
    workflow_function = load_asset(assets.workflow_function, AssetType.TRANSPORTABLE)
    workflow_function_string = load_asset(assets.workflow_function_string, AssetType.TEXT)
    doc = load_asset(assets.doc, AssetType.TEXT)
    named_args = load_asset(assets.named_args, AssetType.OBJECT)
    named_kwargs = load_asset(assets.named_kwargs, AssetType.OBJECT)
    cova_imports = load_asset(assets.cova_imports, AssetType.OBJECT)
    lattice_imports = load_asset(assets.lattice_imports, AssetType.OBJECT)
    deps = load_asset(assets.deps, AssetType.OBJECT)
    call_before = load_asset(assets.call_before, AssetType.OBJECT)
    call_after = load_asset(assets.call_after, AssetType.OBJECT)
    return {
        "workflow_function": workflow_function,
        "workflow_function_string": workflow_function_string,
        "__doc__": doc,
        "named_args": named_args,
        "named_kwargs": named_kwargs,
        "cova_imports": cova_imports,
        "lattice_imports": lattice_imports,
        "metadata": {
            "deps": deps,
            "call_before": call_before,
            "call_after": call_after,
        },
    }


def serialize_lattice(lat, storage_path: str) -> LatticeSchema:
    meta = _serialize_lattice_metadata(lat)
    assets = _serialize_lattice_assets(lat, storage_path)

    tg = serialize_transport_graph(lat.transport_graph, storage_path)

    return LatticeSchema(metadata=meta, assets=assets, transport_graph=tg)


def deserialize_lattice(model: LatticeSchema) -> Lattice:
    def dummy_function(x):
        return x

    lat = Lattice(dummy_function)

    attrs = _deserialize_lattice_metadata(model.metadata)
    assets = _deserialize_lattice_assets(model.assets)

    metadata = assets.pop("metadata")
    attrs.update(assets)
    attrs["metadata"].update(metadata)

    tg = deserialize_transport_graph(model.transport_graph)

    attrs["transport_graph"] = tg

    lat.__dict__.update(attrs)

    if lat.named_args:
        lat.args = [v for _, v in lat.named_args.items()]
    if lat.named_kwargs:
        lat.kwargs = lat.named_kwargs

    return lat



# Copyright 2023 Agnostiq Inc.

import os
import platform
import time
import uuid
from typing import Dict, List, Optional
from covalent_cloud.shared.classes.exceptions import CovalentSDKError

import pennylane as qml
import requests

from ...schemas.schemas import QNodeSpecs, QExecutorSpecs, CircuitCreateSchema
from .base_client import BaseQClient
from ....shared.classes.api import QElectronAPI
from ...._serialize.common import _serialize_asset, _deserialize_asset, AssetType

from covalent._shared_files import logger
app_log = logger.app_log

"""QClient for a remote QServer"""

qserver_url = os.environ.get("COVALENT_CLOUD_QSERVER_URL", "http://covalent-qserver-dev-nlb-4985aa434c6a01e6.elb.us-east-1.amazonaws.com")

class RemoteQClient(BaseQClient):
    def __init__(self) -> None:

        self.api_client = QElectronAPI(qserver_url, url_prefix="/api/v1")

        # Since the cloud qserver only tracks circuits and not
        # batches, the onus of assembling execution results within a
        # batch falls on the qclient.
        self._circuit_ids_by_batch_id = dict()

    @property
    def selector(self):
        return self.deserialize(self.qserver.selector)

    @selector.setter
    def selector(self, selector_func):
        self.qserver.selector = self.serialize(selector_func)

    @property
    def database(self):
        return self.deserialize(self.qserver.database)

    # This will POST to the /circuits
    def register(self, qscripts, executors, qelectron_info, qnode_specs) -> List[Dict]:
        allowed_executors = []
        # Each executor is a Pydantic model
        for executor in executors:
            attrs = executor.dict()
            name = attrs.get("name", "QiskitExecutor")
            print(f"executor dict: {attrs}")
            allowed_executors.append(QExecutorSpecs(name=name, attributes=attrs))

        app_log.debug(f"allowed_executors: {allowed_executors}")
        print(f"qe_info: {qelectron_info}")

        circuit_name = qelectron_info.name
        circuit_description = qelectron_info.description

        dispatch_id = os.getenv("COVALENT_DISPATCH_ID")
        node_id = int(os.getenv("COVALENT_CURRENT_NODE_ID"))

        request_body = CircuitCreateSchema(
            python_version=platform.python_version(),
            dispatch_id=dispatch_id,
            node_id=node_id,
            circuit_name=circuit_name,
            circuit_description=circuit_description or "",
            qnode_specs=qnode_specs,
            allowed_qexecutors=allowed_executors,
            num_circuits=len(qscripts),
        )
        print(f"create circuit request body: {request_body}")
        res = self.api_client.post("/circuits", request_options={
            "data": request_body.json(),
        })
        return res.json()

    def upload_asset(self, asset: bytes, url: str):
        raise NotImplementedError

    def start(self, circuit_ids: List[str]):
        res = self.api_client.post("/qdispatch", request_options={
            "json": circuit_ids
        })
        return res

   
    def submit(
        self, qscripts: List[qml.tape.QuantumScript], executors, qelectron_info, qnode_specs
    ) -> str:
        '''
        Submit a batch of circuits to the qserver.
        
            Circuit submission consists of three stages:
            
            1. Register the circuits by POSTIng to /circuits. Save the
            returned sequence of circuit_ids.
            
            2. Upload circuits (qscripts) and circuit strings to the URLs returned in
            the previous response.
            
            3. Start the circuits by POSTing to /qdispatch
        '''
        circuit_schemas = self.register(qscripts, executors, qelectron_info, qnode_specs)
        app_log.debug(f"circuit_schemas: {circuit_schemas}")
        circuit_ids = list(map(lambda circuit: circuit.get("circuit_id"), circuit_schemas))

        if len(circuit_ids) == 0:
            raise CovalentSDKError("No circuits were returned by the qserver after submission, no valid batch id available.", "remote_qclient/no-circuits-created")

        # Identify a batch of circuits by the first circuit id in the batch
        batch_id = circuit_schemas[0]["batch_id"]
        self._circuit_ids_by_batch_id[batch_id] = circuit_ids

        # Upload serialized circuits and diagrams to the presigned S3 URLs
        print("Uploading circuits and diagrams")
        for i in range(len(qscripts)):
            circuit_url = circuit_schemas[i]["assets"]["circuit"]["uri"]
            circuit_str_url = circuit_schemas[i]["assets"]["circuit_string"]["uri"]

            # get serialized qscript and circuit diagram string
            ser_circuit = self.serialize(qscripts[i])
            circuit_str = qscripts[i].draw().encode("utf-8")

            # upload assets to S3 using presigned URLs
            requests.put(circuit_url, data=ser_circuit).raise_for_status()
            requests.put(circuit_str_url, data=circuit_str).raise_for_status()


        print(f"Started circuits {circuit_ids}")
        self.start(circuit_ids)

        return batch_id

    def get_remaining_circuit_ids(self, batch_id: str) -> List[str]:
        """Returns circuit_ids for pending results."""
        res = self.api_client.get(f'/circuits/batches/{batch_id}')
        circuits = res.json()
        remaining_circuit_ids = []
        for circuit in circuits:
            if circuit.get("status") not in ["COMPLETED", "FAILED"]:
                remaining_circuit_ids.append(circuit.get("circuit_id"))
        return remaining_circuit_ids

    # This will need to poll the qserver
    def get_results(self, batch_id: str):
        # get all circuits to query for results
        remaining_circuit_ids = self._circuit_ids_by_batch_id[batch_id]
        while len(remaining_circuit_ids) > 0:
            remaining_circuit_ids = self.get_remaining_circuit_ids(batch_id)
            print(f"Waiting for results from batch {batch_id}...")
            time.sleep(5)

        print(f"All circuits in batch {batch_id} have reached a terminal state")

        # Download results
        deserialized_results = []
        for cid in self._circuit_ids_by_batch_id[batch_id]:
            res = self.api_client.get(f'/circuits/{cid}')
            circuit_info = res.json()
            result_uri = circuit_info["assets"]["result"]["uri"]
            r = requests.get(result_uri)
            r.raise_for_status()
            deserialized_results.append(self.deserialize(r.content))

        return deserialized_results

    def serialize(self, obj):
        return _serialize_asset(obj, AssetType.OBJECT)

    def deserialize(self, ser_obj):
        return _deserialize_asset(ser_obj, AssetType.OBJECT)


# Copyright 2023 Agnostiq Inc.


import json
from dataclasses import asdict
from datetime import timedelta
from typing import Union

from pydantic import validator
from pydantic.dataclasses import dataclass
from pydantic.json import pydantic_encoder

executor_plugin_name = "cloud"


@dataclass
class CloudExecutor:
    """
        CloudExecutor represents a configuration for executing a Covalent workflow on the Covalent Cloud.
        This class allows users to configure the resources (such as the number of CPUs, memory, GPUs, and GPU type) and the software environment for a Covalent workflow that will be executed on the Covalent Cloud. The time limit for the workflow execution can also be set.
    
        Attributes:
    
            num_cpus (int, optional): Number of CPUs to be used for the workflow. Defaults to 1.
            memory (int, optional): Amount of memory (in MB) to be used for the workflow. Defaults to 1024.
            num_gpus (int, optional): Number of GPUs to be used for the workflow. Defaults to 0.
            gpu_type (str, optional): Type of GPU to be used for the workflow. Defaults to an empty string.
            env (str, optional): Name of the software environment to be used for the workflow. Defaults to "default".
            time_limit (Union[int, timedelta], optional): Time limit for the workflow execution, in seconds or as a timedelta. Defaults to 60.
    
        Examples:
    
            # create a CloudExecutor with default resource configuration
            # executor = CloudExecutor()
    
            # create a custom CloudExecutor with specified resources and environment
            # executor = CloudExecutor(
            #     num_cpus=4,
            #     memory=2048,
            #     num_gpus=1,
            #     gpu_type="NVIDIA-Tesla-V100",
            #     env="my_custom_env",
            #     time_limit=1800  # 30 minutes
            # )
    
            import covalent as ct
            from covalent_cloud import CloudExecutor
    
            cloud_executor1 = CloudExecutor(num_cpus=1, memory=1024)
            cloud_executor2 = CloudExecutor(num_cpus=2, memory=2048)
            cloud_executor3 = CloudExecutor(num_cpus=1, memory=512)
    
            # Define manageable tasks as electrons with different cloud executors
            @ct.electron(executor=cloud_executor1)
            def add(x, y):
                return x + y
    
            @ct.electron(executor=cloud_executor2)
            def multiply(x, y):
                return x * y
    
            @ct.electron(executor=cloud_executor3)
            def divide(x, y):
                return x / y
    
            # Define the workflow as a lattice
            @ct.lattice
            def workflow(x, y):
                r1 = add(x, y)
                r2 = [multiply(r1, y) for _ in range(4)]
                r3 = [divide(x, value) for value in r2]
                return r3
    
            # Import the Covalent Cloud module
            import covalent_cloud as ctc
    
            # Dispatch the workflow to the Covalent Cloud
            dispatch_id = ctc.dispatch(workflow)(1, 2)
            result = ctc.get_result(dispatch_id, wait=True)
            print(result)
    """

    num_cpus: int = 1
    memory: int = 1024
    num_gpus: int = 0
    gpu_type: str = ""
    env: str = "default"
    time_limit: Union[int, timedelta] = 60

    @property
    def short_name(self) -> str:
        """
        Property which returns the short name
        of the executor used by Covalent for identification.

        Args:
            None
        
        Returns:
            The short name of the executor

        """

        return executor_plugin_name

    # Convert timedelta to seconds
    def __post_init__(self):
        if isinstance(self.time_limit, timedelta):
            self.time_limit = self.time_limit.total_seconds()

    # Validators:
    @validator("num_cpus", "memory", "time_limit")
    def gt_than_zero(cls, v: int) -> int:
        """
        Validator which ensures that the value is greater than 0.

        Args:
            v: The value to validate

        Returns:
            The validated value

        """
        
        if v <= 0:
            raise ValueError(f"{v} must be greater than 0")
        return v

    # Model methods:
    def to_json(self) -> str:
        """
        Return a JSON-serialized string representation of this object.

        Args:
            None
        
        Returns:
            The JSON-serialized string representation of this object.
        
        """

        return json.dumps(self, default=pydantic_encoder)

    def to_dict(self) -> dict:
        """
        Return a JSON-serializable dictionary representation of this object.

        Args:
            None
        
        Returns:
            The JSON-serializable dictionary representation of this object.
        """

        return {
            "type": str(self.__class__),
            "short_name": self.short_name,
            "attributes": asdict(self),
        }



# Copyright 2023 Agnostiq Inc.


from typing import Any

import pennylane as qml

from ..middleware.core import middleware
from pennylane.tape import QuantumTape


class QNodeFutureResult:
    """
    A class that stores the `batch_id` of a batch of circuits submitted to the
    middleware. The `result` method can then be called to retrieve the results.

    Attributes:
        device: The Pennylane device used by the original QNode.
        interface: The interface of the original QNode.
        diff_method: The differentiation method of the original QNode.
        qfunc_output: The return value (measurement definition) of the original QNode.
    """

    def __init__(
        self,
        batch_id: str,
        original_qnode: qml.QNode,
        original_tape: QuantumTape,
    ):
        """
        Initialize a `QNodeFutureResult` instance.

        Args:
            batch_id: A UUID that identifies a batch of circuits submitted to
                the middleware.
        """
        self.batch_id = batch_id

        # Required for batch_transforms and correct output typing.
        self.device = original_qnode.device
        self.interface = original_qnode.interface
        self.diff_method = original_qnode.diff_method
        self.tape = original_tape

        self._result = None

    def result(self) -> Any:
        """
        Retrieve the results for the given `batch_id` from middleware. This method
        is blocking until the results are available.

        Returns:
            The results of the circuit execution.
        """

        if self._result is None:

            # Get raw results from the middleware.
            results = middleware.get_results(self.batch_id)

            # Create a device from the original QNode's device class for correct typing.
            dev = _run_later_device_factory(results, self.device, self.tape)

            # Define a dummy circuit that returns the original QNode's return value.
            @qml.qnode(dev, interface=self.interface, diff_method=self.diff_method)
            def _dummy_circuit():
                return self.tape._qfunc_output

            self._result = _dummy_circuit()

        return self._result


def _run_later_device_factory(
    results: Any,
    original_device: qml.Device,
    original_tape: QuantumTape,
) -> qml.Device:
    """
    Returns an instance of a new class that inherits from the original QNode's
    device class. Inheriting ensures the correct return type, while overriding
    `batch_execute` returns the expected `results` without actually running circuits.
    """

    qml_device_cls = type(original_device)

    class _RunLaterDevice(qml_device_cls):

        def batch_execute(self, circuits):
            """
            Override to return expected result.
            """
            n_circuits = len(circuits)

            if n_circuits > 1 or len(results) > 1:
                return [results] * n_circuits
            return results
        
        def batch_transform(self, circuit):
            """
            Override circuit with original measurements for correct `batch_fn`
            choice during execution.
            """
            circuit = QuantumTape(
                ops=circuit.operations,
                measurements=original_tape.measurements,
            )
            return original_device.batch_transform(circuit)

    wires = original_device.num_wires
    return _RunLaterDevice(wires=wires, shots=1)



# Copyright 2023 Agnostiq Inc.

from covalent_cloud.shared.classes.api import DispatcherAPI


def get_client() -> DispatcherAPI:
    """
    A factory / getter method for the Dispatcher API client.

    Args:
        None

    Returns:
        An instance of `DispatcherAPI` client.

    """
    
    return DispatcherAPI()



# Copyright 2023 Agnostiq Inc.

from pydantic import BaseModel
from typing import Dict, List, Optional, Any
import uuid

class QNodeSpecs(BaseModel):
    gate_sizes: Dict[str, int]
    gate_types: Dict[str, int]
    num_operations: int
    num_observables: int
    num_diagonalizing_gates: int
    num_used_wires: int
    depth: int
    num_trainable_params: int = None
    num_device_wires: int
    device_name: str
    diff_method: Optional[str]
    expansion_strategy: str
    gradient_options: Dict[str, int]
    interface: Optional[str]
    gradient_fn: Optional[str]
    num_gradient_executions: Any = 0
    num_parameter_shift_executions: int = None
    
class AssetMetadata(BaseModel):
    uri: str
    size: int


class CircuitAssets(BaseModel):
    circuit: Optional[AssetMetadata] = None
    circuit_string: Optional[AssetMetadata] = None
    result: Optional[AssetMetadata] = None
    result_string: Optional[AssetMetadata] = None


class QExecutorSpecs(BaseModel):
    name: str
    attributes: Dict


# Request body for `/register`
class CircuitCreateSchema(BaseModel):
    """Request body for the /register endpoint"""

    python_version: str
    dispatch_id: uuid.UUID
    node_id: int
    circuit_name: str
    circuit_description: str
    qnode_specs: QNodeSpecs

    # Clarify
    allowed_qexecutors: List[QExecutorSpecs]

    # Number of circuits in the batch
    num_circuits: int

    # Object store URIs -- assigned by the server
    assets: Optional[CircuitAssets] = None

    # Unique ID of a circuit -- assigned by the server
    circuit_id: Optional[str] = None



# Copyright 2023 Agnostiq Inc.


from typing import List

import pennylane as qml

from .qclients import RemoteQClient


class MiddleWare:

    def __init__(self) -> None:
        self.qclient = RemoteQClient()

    def __new__(cls):
        # Making this a singleton class
        if not hasattr(cls, 'instance'):
            cls.instance = super(MiddleWare, cls).__new__(cls)
        return cls.instance

    # The following attributes are properties
    # because the qclient might change over time
    # and every time it gets changed, we shouldn't
    # have to set things like:
    # self.database = self.qclient.database
    # Thus, we override the access of these attributes
    # and return/set these dynamically depending upon
    # what the qclient is at that point in time.

    @property
    def selector(self):
        return self.qclient.selector

    @selector.setter
    def selector(self, selector_func):
        self.qclient.selector = selector_func

    @property
    def database(self):
        return self.qclient.database

    def run_circuits_async(
            self, qscripts: List[qml.tape.qscript.QuantumScript], executors, qelectron_info, qnode_specs):
        return self.qclient.submit(qscripts, executors, qelectron_info, qnode_specs)

    def get_results(self, batch_id):
        return self.qclient.get_results(batch_id)


middleware = MiddleWare()



# Copyright 2023 Agnostiq Inc.


"""Module for Covalent Cloud dispatching and related functionalities."""

import sys
import tempfile
from concurrent.futures import ThreadPoolExecutor, wait
from copy import deepcopy
from functools import wraps
from pathlib import Path
from typing import Callable, List, Optional

import requests
from covalent._results_manager.result import Result
from covalent._shared_files.defaults import parameter_prefix
from covalent._workflow.lattice import Lattice
from rich.progress import BarColumn, MofNCompleteColumn, Progress, TaskID

from .._serialize.result import merge_response_manifest, serialize_result, strip_local_uris
from ..shared.classes.api import APIClient
from ..shared.classes.exceptions import CovalentAPIKeyError, CovalentSDKError
from ..shared.classes.settings import Settings, settings
from ..shared.schemas.asset import AssetSchema
from ..shared.schemas.result import ResultSchema

_dispatch_executor = ThreadPoolExecutor()

_VALID_EXECUTORS = {"cloud"}

dispatch_cache_dir = Path(settings.dispatch_cache_dir)
dispatch_cache_dir.mkdir(parents=True, exist_ok=True)


class AssetUploadThreadFailure(CovalentSDKError):
    message: str = "One or more asset upload threads did not finish execution."
    code: str = "dispatch/asset-upload-thread/fail"

    def __init__(self) -> None:
        super().__init__(self.message, self.code)


class AssetUploadException(CovalentSDKError):
    message: str = "One or more asset upload raised exceptions."
    code: str = "dispatch/asset-upload/exception"

    def __init__(self, msg: str = ""):
        super().__init__(message=msg, code=self.code)


def validate_executors(lat: Lattice) -> bool:

    # Check lattice default executor and workflow_executor

    valid_lattice_executors = True
    valid_electron_executors = True

    if lat.metadata["executor"] not in _VALID_EXECUTORS:
        app_log.debug(f"Found illegal lattice executor: {lat.metadata['executor']}")
        valid_lattice_executors = False
    if lat.metadata["workflow_executor"] not in _VALID_EXECUTORS:
        app_log.debug(f"Found illegal lattice workflow executor: {lat.metadata['workflow_executor']}")
        valid_lattice_executors = False

    tg = lat.transport_graph

    for i in tg._graph.nodes:
        name = tg.get_node_value(i, "name")

        if name.startswith(parameter_prefix):
            continue

        metadata = tg.get_node_value(i, "metadata")
        if metadata["executor"] not in _VALID_EXECUTORS:
            app_log.debug(f"Found illegal electron executor: {metadata['executor']} in node {i}")
            valid_electron_executors = False
            break

    return valid_lattice_executors and valid_electron_executors


def inject_parameter_outputs(lat: Lattice):
    # Hack to pre-compute the output of parameter nodes. This will be
    # no longer needed once OS develop does the same during
    # `build_graph`.

    tg = lat.transport_graph
    if not tg._graph.nodes:
        return

    for node_id in tg._graph.nodes:
        name = tg.get_node_value(node_id, "name")
        if name.startswith(parameter_prefix):
            value = tg.get_node_value(node_id, "value")
            tg.set_node_value(node_id, "output", value)


# For multistage dispatches


def register(
    orig_lattice: Lattice,
    settings: Settings = settings,
) -> Callable:
    """
    Wrapping the dispatching functionality to allow input passing
    and server address specification.

    Afterwards, send the lattice to the dispatcher server and return
    the assigned dispatch id.

    Args:
        orig_lattice: The lattice/workflow to send to the dispatcher server.
        dispatcher_addr: The address of the dispatcher server.  If None then then defaults to the address set in Covalent's config.

    Returns:
        Wrapper function which takes the inputs of the workflow as arguments
    """

    dispatcher_addr = settings.dispatcher_uri

    @wraps(orig_lattice)
    def wrapper(*args, **kwargs) -> str:
        """
        Send the lattice to the dispatcher server and return
        the assigned dispatch id.

        Args:
            *args: The inputs of the workflow.
            **kwargs: The keyword arguments of the workflow.

        Returns:
            The dispatch id of the workflow.
        """

        lattice = deepcopy(orig_lattice)

        lattice.build_graph(*args, **kwargs)

        # Temporary workaround until OS develop also
        # does this during `build_graph.

        inject_parameter_outputs(lattice)

        with tempfile.TemporaryDirectory() as tmp_dir:
            manifest = prepare_manifest(lattice, tmp_dir)
            return_manifest = register_manifest(manifest, settings)
            dispatch_id = return_manifest.metadata.dispatch_id

            path = dispatch_cache_dir / f"{dispatch_id}"

            with open(path, "w") as f:
                f.write(manifest.json())

            upload_assets(manifest)

            return dispatch_id

    return wrapper


def start(
    dispatch_id: str,
    settings: Settings = settings,
) -> Callable:
    """
    Wrapping the dispatching functionality to allow input passing
    and server address specification.

    Afterwards, send the lattice to the dispatcher server and return
    the assigned dispatch id.

    Args:
        orig_lattice: The lattice/workflow to send to the dispatcher server.
        dispatcher_addr: The address of the dispatcher server.  If None then then defaults to the address set in Covalent's config.

    Returns:
        Wrapper function which takes the inputs of the workflow as arguments
    """

    dispatcher_addr = settings.dispatcher_uri

    test_url = f"{dispatcher_addr}/api/v1/dispatchv2/start/{dispatch_id}"

    client = APIClient(host_uri=dispatcher_addr, settings=settings)
    endpoint = f"/api/v1/dispatchv2/start/{dispatch_id}"

    try:
        r = client.put(endpoint)
    except requests.exceptions.HTTPError as e:
        print(e.response.text, file=sys.stderr)
        raise e
    return r.content.decode("utf-8").strip().replace('"', "")


def prepare_manifest(lattice, storage_path) -> ResultSchema:
    """Prepare a built-out lattice for submission"""

    result_object = Result(lattice)
    return serialize_result(result_object, storage_path)


def register_manifest(
    manifest: ResultSchema,
    settings: Settings = settings,
    parent_dispatch_id: Optional[str] = None,
    push_assets: bool = True,
) -> ResultSchema:
    """Submits a manifest for registration.

    Returns:
        Dictionary representation of manifest with asset remote_uris filled in

    Side effect:
        If push_assets is False, the server will
        automatically pull the task assets from the submitted asset URIs.

    Raises:
        CovalentAPIKeyError: If the API key is invalid.

    """
    dispatcher_addr = settings.dispatcher_uri

    stripped = strip_local_uris(manifest) if push_assets else manifest
    client = APIClient(host_uri=dispatcher_addr, settings=settings)
    endpoint = "/api/v1/dispatchv2/register"

    if parent_dispatch_id:
        endpoint = f"{endpoint}/{parent_dispatch_id}"

    try:
        r = client.post(endpoint, request_options={"data": stripped.json()})
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 401 and e.response.json()["code"] == "auth/unauthorized":
            raise CovalentAPIKeyError(
                message="A valid API key is required to register a dispatch.",
                code=e.response.json()["code"],
            ) from e
        else:
            raise

    parsed_resp = ResultSchema.parse_obj(r.json())

    return merge_response_manifest(manifest, parsed_resp)


def upload_assets(manifest: ResultSchema):
    assets = _extract_assets(manifest)
    _upload(assets)


def _extract_assets(manifest: ResultSchema) -> List[AssetSchema]:
    # workflow-level assets
    dispatch_assets = manifest.assets
    assets = [asset for key, asset in dispatch_assets]
    lattice = manifest.lattice
    lattice_assets = lattice.assets
    assets.extend(asset for key, asset in lattice_assets)
    # Node assets
    tg = lattice.transport_graph
    nodes = tg.nodes
    for node in nodes:
        node_assets = node.assets
        assets.extend(asset for key, asset in node_assets)
    return assets


def _upload(assets: List[AssetSchema]) -> None:
    """Upload assets to remote storage.

    Args:
        assets: List of AssetSchema objects to upload.

    Raises:
        RuntimeError: If any of the assets fail to upload.

    Returns:
        None

    """
    total_assets = len(assets)
    _upload_futures = []

    with Progress(
        "[progress.description]{task.description}",
        BarColumn(),
        "[progress.percentage]{task.percentage:>4.1f}%",
        MofNCompleteColumn(),
    ) as progress:
        task = progress.add_task("[green]Uploading assets...", total=total_assets)

        for asset in assets:
            fut = _dispatch_executor.submit(
                _upload_asset, asset.uri, asset.remote_uri, task, progress
            )
            _upload_futures.append(fut)

        done, _ = wait(_upload_futures)

        if len(done) < total_assets:
            raise AssetUploadThreadFailure

        _exceptions = []
        for fut in done:
            if ex := fut.exception(timeout=0.1):
                _exceptions.append(ex)

        if _exceptions:
            msg = f"Failed to upload {len(_exceptions)} out of {total_assets} due to raised exceptions."
            print(msg, file=sys.stderr)
            raise AssetUploadException(msg)


def _upload_asset(local_uri: str, remote_uri: str, task: TaskID, progress: Progress) -> None:
    """Upload a single asset to remote storage.

    Args:
        local_uri: Local URI of the asset to upload.
        remote_uri: Remote URI to upload the asset to.
        task: Task ID of the progress bar task.
        progress: Progress bar object.

    Raises:
        requests.exceptions.HTTPError: If the upload fails.

    Returns:
        None

    """
    scheme_prefix = "file://"
    if local_uri.startswith(scheme_prefix):
        local_path = local_uri[len(scheme_prefix) :]
    else:
        local_path = local_uri

    with open(local_path, "rb") as f:
        files = {"asset_file": f}
        r = requests.post(remote_uri, files=files)
        r.raise_for_status()

    if r.status_code == requests.codes.ok:
        progress.advance(task, advance=1)
        progress.refresh()



# Copyright 2023 Agnostiq Inc.

from typing import Optional, Dict
from pydantic import BaseModel
from .base import CloudQExecutor


class Options(BaseModel):
    """
    Options for the QiskitExecutor
    """

    optimization_level: int = 3
    resilience_level: int = 3
    max_execution_time: Optional[int] = None
    transpilation: Optional[Dict] = None
    resilience: Optional[Dict] = None
    execution: Optional[Dict] = None
    environment: Optional[Dict] = None
    simulator: Optional[Dict] = None


class QiskitExecutor(CloudQExecutor):
    """
    QiskitExecutor represents the configuration to use when executing the QElectron
    on the Qiskit backend.
    """

    backend: str
    shots: Optional[int] = -1 # -1 means use the `shots` value provided in the QNode's device
    single_job: bool = False
    max_execution_time: int = None
    options: Options = None

    def validate_attrs(self):
        """
        Validate the attributes of the executor and whether they are compatible with the QNode's device.
        Currently validates the `shots` attribute.
        """
        if self.shots <= 0:
            raise ValueError(f"QiskitExecutor does not support setting shots <= 0 current value is {self.shots}. Please check the QNode's device or QiskitExecutor's shot.")



# Copyright 2023 Agnostiq Inc.

"""Models for /api/v1/resultv2 endpoints"""

from typing import List

from pydantic import BaseModel

from .edge import EdgeSchema
from .electron import ElectronSchema


class TransportGraphSchema(BaseModel):
    nodes: List[ElectronSchema]
    links: List[EdgeSchema]
